SystemOrganization addCategory: #SPN!
SystemOrganization addCategory: 'SPN-Tests'!

TestCase subclass: #SPNTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SPN-Tests'!

!SPNTest commentStamp: '<historical>' prior: 0!
A SPNTest is a test class for testing the behavior of SPN!

!SPNTest methodsFor: 'tests-accessing' stamp: 'Lolita 5/20/2017 00:10'!
testEncrypt
	self
		assert: (SPN new encrypt: '0010011010110111')
		equals: '1011110011010110'! !

!SPNTest methodsFor: 'tests-accessing' stamp: 'Lolita 5/20/2017 00:09'!
testLinear_cryptanalysisWith
	self
		assert: (SPN new linear_cryptanalysis: (SPN generatePairs: 8000) with: 8000 )
		equals: '01101111'! !

Object subclass: #SPN
	instanceVariableNames: 'substitution_fct permutation_fct keys'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SPN'!

!SPN commentStamp: 'Lolita 5/9/2017 18:46' prior: 0!
I'm an iterated encryption algorithm. I'm constructed for strings consisting of 0s and 1s, and which size is 16.
I can encrypt strings with the method encrypt.

You can modify my substitution and permutation functions with the methods substitution_fct: and permutation_fct:.

Example:
SPN new encrypt: '0010011010110111'
!

!SPN class methodsFor: 'accessing' stamp: 'Lolita 5/19/2017 10:03'!
generatePairs: anInteger
	"Answers a dictionnary containing anInteger pairs of (clair string - encrypted string) where the keys are the clair strings"
	| dictionnary clair_string encrypted_string spn|
	dictionnary := Dictionary new.
	spn := self new.
	
	1 to: anInteger do: [ :each |
		clair_string := ''.
		1 to: 16 do: [ :each1 | 
			clair_string := clair_string , ((0 to: 1) atRandom) greaseString 
		].
		encrypted_string := spn encrypt: clair_string.
		dictionnary at: clair_string  put: encrypted_string.
	].
	^ dictionnary ! !

!SPN methodsFor: 'accessing' stamp: 'Lolita 5/20/2017 00:02'!
encrypt: aString
	"answers a string encrypted by the SPN algorithm"

	| int_string substitution_s key_incorporated_s encrypted_s|
	int_string := aString.
	Transcript cr; show: 'Intermediate string (w0): '; show: int_string.
	
	1 to: 3 do: [ :each | 
		key_incorporated_s := self key: (self keys at: each) with: int_string length: 16.
		Transcript cr; show: 'KeyIncorporated string (u',  each greaseString , '): '; show: key_incorporated_s.
		
		substitution_s := ''.
		1 to: 4 do: [ :each1 | 
			substitution_s := substitution_s , (self substitution_fct
				at: ( key_incorporated_s copyFrom: ((4 * (each1 - 1)) + 1) to: 4 * each1 ) )
				 ].
		Transcript cr; show: 'Subtitution string (v', each greaseString, '): '; show: substitution_s.	
			
		int_string := '0000000000000000'.
		1 to: 16 do: [ :each2 | int_string at: each2 put: (substitution_s at: (permutation_fct at: each2))].
		Transcript cr; show: 'Intermediate string (w', each greaseString , '): '; show: int_string.
	].
	
	key_incorporated_s := self key: (self keys at: 4) with: int_string length: 16.
	Transcript cr; show: 'KeyIncorporated string (u4): '; show: key_incorporated_s.
	
	substitution_s := ''.
	1 to: 4 do: [ :each1 | 
			substitution_s := substitution_s , (self substitution_fct
				at: (key_incorporated_s copyFrom: ((4 * (each1 - 1)) + 1) to: 4 * each1 ) ) ].
	Transcript cr; show: 'Subtitution string (v4): '; show: substitution_s.	
	
	encrypted_s := self key: (self keys at: 5) with: substitution_s length: 16.
	Transcript cr; cr; show: 'Encrypted string (y): '; show: encrypted_s.
	
	^ encrypted_s 
	
! !

!SPN methodsFor: 'as yet unclassified' stamp: 'Lolita 5/19/2017 10:41'!
hexToBinTable

	| hexToBin |
	
	hexToBin := Dictionary new.
	hexToBin at: 0 put: '0000';
		at: 1 put: '0001';
		at: 2 put: '0010';
		at: 3 put: '0011';
		at: 4 put: '0100';
		at: 5 put: '0101';
		at: 6 put: '0110';
		at: 7 put: '0111';
		at: 8 put: '1000';
		at: 9 put: '1001';
		at: 10 put: '1010';
		at: 11 put: '1011';
		at: 12 put: '1100';
		at: 13 put: '1101';
		at: 14 put: '1110';
		at: 15 put: '1111'.
	^ hexToBin ! !

!SPN methodsFor: 'initialization' stamp: 'Lolita 5/8/2017 20:26'!
initialize
	substitution_fct := Dictionary new.
	substitution_fct
		at: '0000' put: '1110';
		at: '0001' put: '0100';
		at: '0010' put: '1101';
		at: '0011' put: '0001';
		at: '0100' put: '0010';
		at: '0101' put: '1111';
		at: '0110' put: '1011';
		at: '0111' put: '1000';
		at: '1000' put: '0011';
		at: '1001' put: '1010';
		at: '1010' put: '0110';
		at: '1011' put: '1100';
		at: '1100' put: '0101';
		at: '1101' put: '1001';
		at: '1110' put: '0000';
		at: '1111' put: '0111'.
	permutation_fct := Dictionary new.
	permutation_fct
		at: 1 put: 1;
		at: 2 put: 5;
		at: 3 put: 9;
		at: 4 put: 13;
		at: 5 put: 2;
		at: 6 put: 6;
		at: 7 put: 10;
		at: 8 put: 14;
		at: 9 put: 3;
		at: 10 put: 7;
		at: 11 put: 11;
		at: 12 put: 15;
		at: 13 put: 4;
		at: 14 put: 8;
		at: 15 put: 12;
		at: 16 put: 16.
	keys := OrderedCollection new.
	keys
		add: '0011101010010100';
		add: '1010100101001101';
		add: '1001010011010110';
		add: '0100110101100011';
		add: '1101011000111111'! !

!SPN methodsFor: 'as yet unclassified' stamp: 'Lolita 5/19/2017 11:21'!
key: aKey with: aString length: anInteger
	"Returns the string consisting at each character in the XOR of the key and the given string"
	
	| result |
	result := String new: anInteger.
	1 to: anInteger do: [ :each |
		( aKey at: each )	= ( aString at: each ) ifTrue: [ result at: each put: $0 ] ifFalse: [ result at: each put: $1 ]
	].
	
	^ result ! !

!SPN methodsFor: 'accessing' stamp: 'Lolita 5/9/2017 04:50'!
keys 
	^ keys! !

!SPN methodsFor: 'accessing' stamp: 'Lolita 5/19/2017 12:34'!
linear_cryptanalysis: aDictionnary with: numberOfElements
	"Answers the best candidate sub-key"
	
	| hexToBin keyPart1is keyPart2is counterKeys u4_2 u4_4 z max maxKey|
	counterKeys := Matrix new: 16 element: 0.
	hexToBin := self hexToBinTable.
	
	"For each word in my dictionnary"
	aDictionnary associationsDo: [ :each |
		0 to: 15 do: [ :keyPart1 |
			0 to: 15 	do: [ :keyPart2 |
				keyPart1is := hexToBin at: keyPart1.
				keyPart2is := hexToBin at: keyPart2.
				u4_2 := self substitution_fct keyAtValue: (self key: keyPart1is with: (each value copyFrom: 5 to: 8 ) length: 4).
				u4_4 := self substitution_fct keyAtValue: (self key: keyPart2is  with: (each value copyFrom: 13 to: 16 ) length: 4).
				z := ((each key at: 5) greaseInteger - 48) + ((each key at: 7) greaseInteger - 48) + ((each key at: 8) greaseInteger - 48) + ((u4_2 at: 2) greaseInteger - 48) + ((u4_2 at: 4) greaseInteger - 48) + ((u4_4 at: 2) greaseInteger - 48) + ((u4_4 at: 4) greaseInteger - 48).
				
				(z \\ 2) = 0 ifTrue: [ counterKeys at: keyPart1 + 1 at: keyPart2 + 1 put: (counterKeys at: keyPart1 + 1 at: keyPart2 + 1) + 1 ]
			]
		]			
	].
	"We can now select the best key"
	max := -1.
	0 to: 15 do: [ :keyPart1 |
		0 to: 15 	do: [ :keyPart2 |
			counterKeys at: keyPart1 + 1 at: keyPart2 + 1 put: (counterKeys at: keyPart1 + 1 at: keyPart2 + 1) - (numberOfElements / 2).
			(counterKeys at: keyPart1 + 1 at: keyPart2 + 1) > max
			ifTrue: [ max := (counterKeys at: keyPart1 + 1 at: keyPart2 + 1).
						 maxKey := (hexToBin at: keyPart1), (hexToBin at: keyPart2) ]
		]
	].
	^ maxKey ! !

!SPN methodsFor: 'as yet unclassified' stamp: 'Lolita 5/19/2017 10:41'!
nLValuesTable
	"Answers a table containing at each cell (i,j) the value Nl(i,j)"
	|hexToBin table ais bis nl vector|
	hexToBin := self hexToBinTable.
	table := Matrix new: 16.
	0 to: 15 do: [ :a |
		0 to: 15 	do: [ :b |
			ais := hexToBin at: a.
			bis := hexToBin at: b.
			nl := 0.
			substitution_fct associationsDo: [ :each | 
				vector := self substitution: (each key, each value) and: (ais, bis).
				((self xor: vector) = 0) ifTrue: [nl := nl +1].
			].
			table at: a + 1 at: b + 1 put: nl.
		]
	].
	^ table

! !

!SPN methodsFor: 'accessing' stamp: 'Lolita 5/9/2017 04:51'!
permutation_fct
	^ permutation_fct ! !

!SPN methodsFor: 'accessing' stamp: 'Lolita 5/9/2017 07:17'!
permutation_fct: aDictionnary
	permutation_fct := aDictionnary ! !

!SPN methodsFor: 'as yet unclassified' stamp: 'Lolita 5/18/2017 08:08'!
substitution: xy and: ab
	| result |
	result := Array new: 8.
	1 to: 8 do: [ :i | 
		((xy at: i) = $1 and: (ab at: i) = $1) 
		ifTrue: [ result at: i put: 1] 
		ifFalse: [ result at: i put: 0]
	].
	^ result

! !

!SPN methodsFor: 'accessing' stamp: 'Lolita 5/9/2017 04:50'!
substitution_fct
	^ substitution_fct ! !

!SPN methodsFor: 'accessing' stamp: 'Lolita 5/9/2017 07:17'!
substitution_fct: aDictionnary
	substitution_fct := aDictionnary ! !

!SPN methodsFor: 'as yet unclassified' stamp: 'Lolita 5/18/2017 08:19'!
xor: vector
	| sum |
	sum := 0.
	vector do: [ :each | sum := sum + each ].
	^ sum \\ 2
	

! !
